from pydispix import churches  # noqa: F401: F401
from pydispix.autodraw import AutoDrawer  # noqa: F401
from pydispix.canvas import Canvas, Pixel  # noqa: F401
from pydispix.church import ChurchClient  # noqa: F401
from pydispix.client import Client  # noqa: F401
from pydispix.color import Color, Colour, parse_color, parse_colour  # noqa: F401
from pydispix.log import setup_logging
from pydispix.multiplexing import MultiAutoDrawer, MultiClient  # noqa: F401
from pydispix.ratelimits import RateLimitedEndpoint, RateLimiter  # noqa: F401

setup_logging()



import logging
import os
import sys

import colorama


class ColoredFormatter(logging.Formatter):
    COLORS = {
        "WARNING": colorama.Fore.YELLOW,
        "INFO": colorama.Fore.WHITE,
        "DEBUG": colorama.Fore.BLUE,
        "CRITICAL": f"{colorama.Style.BRIGHT}{colorama.Fore.RED}",
        "ERROR": f"{colorama.Style.DIM}{colorama.Fore.RED}",
    }
    RESET_SEQ = f"{colorama.Style.RESET_ALL}{colorama.Fore.RESET}"

    def __init__(self, fmt: str, *args, use_color: bool = True, **kwargs) -> None:
        super().__init__(fmt, *args, **kwargs)
        self.use_color = use_color

    def format(self, record: logging.LogRecord):
        levelname = record.levelname
        if self.use_color and levelname in self.COLORS:
            levelname_color = self.COLORS[levelname] + levelname + self.RESET_SEQ
            record.levelname = levelname_color
        return super().format(record)


def setup_logging():
    debug_mode = 'DEBUG' in os.environ
    log_format = ColoredFormatter(
        f"{colorama.Fore.GREEN}%(asctime)s {colorama.Fore.RESET} | "
        f"{colorama.Style.BRIGHT} %(name)s {colorama.Style.RESET_ALL}   | "
        "%(levelname)s  | %(message)s"
    )
    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(log_format)

    logger = logging.getLogger("pydispix")
    logger.setLevel(logging.DEBUG if debug_mode else logging.INFO)
    logger.addHandler(stream_handler)



import logging
import random
import re
import time
from dataclasses import dataclass
from json.decoder import JSONDecodeError

import requests

from pydispix.church import ChurchClient, ChurchTask
from pydispix.errors import RateLimitBreached, get_response_result

logger = logging.getLogger("pydispix")

SQLITE_CHURCH = "https://decorator-factory.su"
RICK_CHURCH = "https://pixel-tasks.scoder12.repl.co/api"


@dataclass
class RickChurchTask(ChurchTask):
    project_title: str
    start: float


@dataclass
class SQLiteChurchTask(ChurchTask):
    id: int
    issued_by: str


class RickChurchClient(ChurchClient):
    """Church Client designed to work specifically with rick church"""

    def __init__(
        self,
        pixel_api_token: str,
        church_token: str,
        base_church_url: str = RICK_CHURCH,
        *args,
        **kwargs
    ):
        super().__init__(pixel_api_token, church_token, base_church_url, *args, **kwargs)

    def get_task(self, repeat_delay: int = 2) -> RickChurchTask:
        url = self.resolve_church_endpoint("get_task")
        while True:
            response = self.make_request("GET", url, params={"key": self.church_token}).json()

            if response["task"] is None:
                logger.info(f"Church doesn't currently have any aviable tasks, waiting {repeat_delay}s")
                time.sleep(repeat_delay)
                continue
            return RickChurchTask(**response["task"])

    def submit_task(self, church_task: RickChurchTask, endpoint: str = "submit_task") -> requests.Response:
        url = self.resolve_church_endpoint(endpoint)
        body = {
            'project_title': church_task.project_title,
            'start': church_task.start,
            'x': church_task.x,
            'y': church_task.y,
            'color': church_task.color
        }
        req = self.make_request("POST", url, data=body, params={"key": self.church_token})
        completed_tasks = self.get_personal_stats()["goodTasks"]
        logger.info(f"Task submitted to the church (tasks complete={completed_tasks}")
        return req

    def _handle_church_task_errors(self, exception: Exception) -> None:
        """
        Rick church can raise certain specific errors, handle
        them here or raise them back, if they shouldn't be handled.
        """
        if isinstance(exception, RateLimitBreached):
            try:
                detail: str = get_response_result(exception, "detail", error_on_fail=True)  # type: ignore
            except (UnicodeDecodeError, JSONDecodeError, KeyError):
                # If we can't get the detail, this isn't the exception we're looking for
                return super()._handle_church_task_errors(exception)

            match = re.search(
                r"You have not gotten a task yet or you took more than (\d+) seconds to submit your task",
                detail
            )
            if not match:
                # If the detail isn't matching, this isn't an exception from the rick church
                return super()._handle_church_task_errors(exception)

            # Log the exception and proceed cleanly
            logger.warning(f"Church task failed, task disassigned, submitting took over {match.groups()[0]} seconds")
        elif isinstance(exception, requests.HTTPError):
            try:
                detail: str = get_response_result(exception, "detail", error_on_fail=True)  # type: ignore - if it's not str, we handle it
            except (UnicodeDecodeError, JSONDecodeError, KeyError):
                # If we can't get the detail, this isn't the exception we're looking for
                return super()._handle_church_task_errors(exception)

            if exception.response.status_code == 409:
                if detail != "This is not the task you were assigned":
                    # If the detail isn't matching, this isn't an exception from the rick church
                    return super()._handle_church_task_errors(exception)

                # Log the exception and proceed cleanly
                logger.warning("Church task failed, this task already got reassigned to somebody else.")
            elif exception.response.status_code == 400:
                msg = (
                    "You did not complete this task properly, or it was fixed before the server could verify it. "
                    "You have not been credited for this task."
                )
                if detail != msg:
                    # If the detail isn't matching, this isn't an exception from the rick church
                    return super()._handle_church_task_errors(exception)

                # Log the exception and proceed cleanly
                logger.warning("Church task failed, check failed, someone has overwritten the pixel before we could submit it.")
        elif isinstance(exception, requests.exceptions.SSLError):
            url = exception.request.url
            if not url.startswith(self.base_church_url):
                # This error doesn't come from church of rick URL
                return super()._handle_church_task_errors(exception)

            # Log the exception and proceed cleanly
            logger.warning("Church task failed, SSL Error: Church of rick's SSL certificate wasn't valid. For some reason this sometimes occurs.")
        else:
            # If we didn't find a rich church specific exception,
            # call the super's implementation of this, there could
            # be some other common errors
            return super()._handle_church_task_errors(exception)

    # region: Add some misc endpoints which Church of Rick provides

    def get_personal_stats(self):
        """Get personal stats."""
        url = self.resolve_church_endpoint("user/stats")
        return self.make_request("GET", url, params={"key": self.church_token}).json()

    def get_church_stats(self):
        """Get church stats."""
        url = self.resolve_church_endpoint("overall_stats")
        return self.make_request("GET", url).json()

    def get_leaderboard(self) -> list:
        """Get church leaderboard."""
        url = self.resolve_church_endpoint("leaderboard")
        return self.make_request("GET", url).json()["leaderboard"]

    def get_uptime(self) -> float:
        """Uptime of the church of rick."""
        url = self.resolve_church_endpoint("leaderboard")
        return float(self.make_request("GET", url).json()["uptime"])

    def get_projects(self) -> list:
        """Get project data from the church."""
        url = self.resolve_church_endpoint("projects/stats")
        return self.make_request("GET", url).json()

    # endregion


class SQLiteChurchClient(ChurchClient):
    """Church Client designed to work specifically with rick church"""

    def __init__(
            self,
            pixel_api_token: str,
            base_church_url: str = SQLITE_CHURCH,
            *args,
            **kwargs
    ):
        # SQLite Church API is open for everyone, it doesn't need a token
        church_token = ""
        super().__init__(pixel_api_token, church_token, base_church_url, *args, **kwargs)

    def get_task(self, endpoint: str = "tasks", repeat_delay: int = 2) -> SQLiteChurchTask:
        url = self.resolve_church_endpoint(endpoint)
        while True:
            response = self.make_request("GET", url).json()

            if len(response) == 0:
                logger.info(f"Church doesn't currently have any aviable tasks, waiting {repeat_delay}s")
                time.sleep(repeat_delay)
                continue
            # SQLite church returns a list of aviable tasks to complete, it doesn't assign
            # specific tasks to members, since there is no unique API key. Best we can do is
            # Therefore to pick a task randomly from this list
            task = random.choice(response)
            return SQLiteChurchTask(**task)

    def submit_task(self, church_task: SQLiteChurchTask, endpoint: str = "submit_task") -> requests.Response:
        url = self.resolve_church_endpoint(endpoint)
        body = {"task_id": church_task.id}
        req = self.make_request("POST", url, data=body)
        logger.info("Task submitted to the church")
        return req



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydispix']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.4,<0.5.0',
 'matplotlib>=3.4.2,<3.5.0',
 'pillow>=8.2.0,<8.3.0',
 'requests>=2.25.1,<2.26.0']

setup_kwargs = {
    'name': 'pydispix',
    'version': '1.0.1',
    'description': "API wrapper for python-discord's pixels.",
    'long_description': None,
    'author': 'ItsDrike',
    'author_email': 'itsdrikeofficial@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<3.9.0',
}


setup(**setup_kwargs)



"""Tool for automatically drawing images."""
import logging
import time
from typing import Iterator, List, Tuple

import PIL.Image

from pydispix.canvas import Canvas, Pixel
from pydispix.client import Client
from pydispix.errors import OutOfBoundaries

logger = logging.getLogger('pydispix')


class AutoDrawer:
    """Tool for automatically drawing images."""

    def __init__(
        self,
        client: Client,
        x: int, y: int,
        grid: List[List[Pixel]]
    ):
        """Store the plan."""
        self.client = client
        self.grid = grid
        # Top left coords.
        self.x0 = x
        self.y0 = y
        # Bottom right coords.
        self.x1 = x + len(self.grid[0])
        self.y1 = y + len(self.grid)

        # Make sure we're within canvas boundaries
        canvas_width, canvas_height = canvas_size = self.client.get_dimensions()
        image_width, image_height = image_size = len(self.grid[0]), len(self.grid)
        if image_width > canvas_width or image_height > canvas_height:
            raise OutOfBoundaries(f"Can't draw picture bigger than the canvas ({image_size} > {canvas_size})")

    @classmethod
    def load_image(
        cls,
        client: Client,
        xy: Tuple[int, int],
        image: PIL.Image.Image,
        scale: float = 1
    ) -> 'AutoDrawer':
        """Draw from the pixels of an image."""
        if image.mode == 'RGBA':
            new_image = PIL.Image.new('RGB', image.size)
            new_image.paste(image, mask=image)
            image = new_image

        width = round(image.width * scale)
        height = round(image.height * scale)
        resized = image.resize((width, height), PIL.Image.BILINEAR)
        data = list(resized.getdata())
        grid = [
            [Pixel(*pixel) for pixel in data[start:start + width]]
            for start in range(0, len(data), width)
        ]
        return cls(client, *xy, grid)

    def _iter_coords(self) -> Iterator[Tuple[int, int]]:
        """Iterate over the coordinates of the image."""
        for x in range(self.x0, self.x1):
            for y in range(self.y0, self.y1):
                yield x, y

    def draw_pixel(self, canvas: Canvas, x: int, y: int, show_progress: bool = True) -> bool:
        """
        Draw a pixel if not already drawn.

        Returns True if the pixel was not already drawn.
        """
        color = self.grid[y - self.y0][x - self.x0]
        if canvas[x, y] == color:
            logger.debug(f'Skipping already correct pixel at {x}, {y}.')
            return False
        self.client.put_pixel(x, y, color, show_progress=show_progress)
        return True

    def draw(self, guard: bool = False, guard_delay: int = 5, show_progress: bool = True):
        """Draw the pixels of the image, attempting each pixel max. once."""
        canvas = self.client.get_canvas()
        while True:
            for x, y in self._iter_coords():
                if self.draw_pixel(canvas, x, y, show_progress=show_progress):
                    canvas = self.client.get_canvas()

            if not guard:
                # Check this here, to act as do-while,
                # (always run first time, only continue if this is met)
                break
            # When we're guarding we need to update canvas even if no pixel was drawn
            # because otherwise we'd be looping over same non-updated canvas forever
            # since this looping with no changes takes a long time, we should also sleep
            # to avoid needless cpu usage
            time.sleep(guard_delay)
            canvas = self.client.get_canvas()
